@@ -44,11 +44,16 @@ module Database.Esqueleto
              , (==.), (>=.), (>.), (<=.), (<.), (!=.), (&&.), (||.)
              , (+.), (-.), (/.), (*.)
              , random_, round_, ceiling_, floor_
-             , min_, max_, sum_, avg_
+             , min_, max_, sum_, avg_, lower_
+             , coalesce, coalesceDefault
              , like, (%), concat_, (++.)
              , subList_select, subList_selectDistinct, valList
              , in_, notIn, exists, notExists
-             , set, (=.), (+=.), (-=.), (*=.), (/=.) )
+             , set, (=.), (+=.), (-=.), (*=.), (/=.)
+             , case_ )
+  , when_
+  , then_
+  , else_
   , from
   , Value(..)
   , unValue
@@ -81,6 +86,7 @@ module Database.Esqueleto
 
     -- * Helpers
   , valkey
+  , valJ
 
     -- * Re-exports
     -- $reexports
@@ -88,6 +94,8 @@ module Database.Esqueleto
   , module Database.Esqueleto.Internal.PersistentImport
   ) where
 
+import Control.Monad.IO.Class (MonadIO)
+import Control.Monad.Trans.Reader (ReaderT)
 import Data.Int (Int64)
 import Database.Esqueleto.Internal.Language
 import Database.Esqueleto.Internal.Sql
@@ -339,7 +347,7 @@ import qualified Database.Persist
 -- In @esqueleto@, we may write the same query above as:
 --
 -- @
--- insertSelect $ from $ \p->
+-- insertSelect $ from $ \\p->
 --  return $ BlogPost \<# \"Group Blog Post\" \<&\> (p ^. PersonId)
 -- @
 --
@@ -371,9 +379,24 @@ import qualified Database.Persist
 
 -- | @valkey i = val (Key (PersistInt64 i))@
 -- (<https://github.com/meteficha/esqueleto/issues/9>).
-valkey :: Esqueleto query expr backend =>
+valkey :: (Esqueleto query expr backend, ToBackendKey SqlBackend entity, PersistField (Key entity)) =>
           Int64 -> expr (Value (Key entity))
-valkey = val . Key . PersistInt64
+valkey = val . toSqlKey
+
+
+-- | @valJ@ is like @val@ but for something that is already a @Value@. The use
+-- case it was written for was, given a @Value@ lift the @Key@ for that @Value@
+-- into the query expression in a type safe way. However, the implementation is
+-- more generic than that so we call it @valJ@.
+--
+-- Its important to note that the input entity and the output entity are
+-- constrained to be the same by the type signature on the function
+-- (<https://github.com/prowdsponsor/esqueleto/pull/69>).
+--
+-- /Since: 1.4.2/
+valJ :: (Esqueleto query expr backend, PersistField (Key entity)) =>
+        Value (Key entity) -> expr (Value (Key entity))
+valJ = val . unValue
 
 
 ----------------------------------------------------------------------
@@ -381,8 +404,8 @@ valkey = val . Key . PersistInt64
 
 -- | Synonym for 'Database.Persist.Store.delete' that does not
 -- clash with @esqueleto@'s 'delete'.
-deleteKey :: ( PersistStore m
-             , PersistMonadBackend m ~ PersistEntityBackend val
+deleteKey :: ( PersistStore (PersistEntityBackend val)
+             , MonadIO m
              , PersistEntity val )
-          => Key val -> m ()
+          => Key val -> ReaderT (PersistEntityBackend val) m ()
 deleteKey = Database.Persist.delete
