@@ -35,6 +35,9 @@ module Database.Esqueleto.Internal.Language
   , PreprocessedFrom
   , From
   , FromPreprocess
+  , when_
+  , then_
+  , else_
   ) where
 
 import Control.Applicative (Applicative(..), (<$>))
@@ -252,6 +255,23 @@ class (Functor query, Applicative query, Monad query) =>
   min_     :: (PersistField a) => expr (Value a) -> expr (Value (Maybe a))
   max_     :: (PersistField a) => expr (Value a) -> expr (Value (Maybe a))
   avg_     :: (PersistField a, PersistField b) => expr (Value a) -> expr (Value (Maybe b))
+  lower_   :: (PersistField a, IsString a) => expr (Value a) -> expr (Value a)
+
+  -- | @COALESCE@ function. Evaluates the arguments in order and
+  -- returns the value of the first non-NULL expression, or NULL
+  -- (Nothing) otherwise. Some RDBMSs (such as SQLite) require
+  -- at least two arguments; please refer to the appropriate
+  -- documentation.
+  --
+  -- /Since: 1.4.3/
+  coalesce :: PersistField a => [expr (Value (Maybe a))] -> expr (Value (Maybe a))
+
+  -- | Like @coalesce@, but takes a non-nullable expression
+  -- placed at the end of the expression list, which guarantees
+  -- a non-NULL result.
+  --
+  -- /Since: 1.4.3/
+  coalesceDefault :: PersistField a => [expr (Value (Maybe a))] -> expr (Value a) -> expr (Value a)
 
   -- | @LIKE@ operator.
   like :: (PersistField s, IsString s) => expr (Value s) -> expr (Value s) -> expr (Value Bool)
@@ -320,6 +340,48 @@ class (Functor query, Applicative query, Monad query) =>
   -- | Apply extra @expr Value@ arguments to a 'PersistField' constructor
   (<&>) :: expr (Insertion (a -> b)) -> expr (Value a) -> expr (Insertion b)
 
+  -- | @CASE@ statement.  For example:
+  --
+  -- @
+  -- select $
+  -- return $
+  -- case_
+  --    [ when_
+  --        (exists $
+  --        from $ \\p -> do
+  --        where_ (p ^. PersonName ==. val "Mike"))
+  --      then_
+  --        (sub_select $
+  --        from $ \\v -> do
+  --        let sub =
+  --                from $ \\c -> do
+  --                where_ (c ^. PersonName ==. val "Mike")
+  --                return (c ^. PersonFavNum)
+  --        where_ (v ^. PersonFavNum >. sub_select sub)
+  --        return $ count (v ^. PersonName) +. val (1 :: Int)) ]
+  --    (else_ $ val (-1))
+  -- @
+  --
+  -- This query is a bit complicated, but basically it checks if a person
+  -- named "Mike" exists, and if that person does, run the subquery to find
+  -- out how many people have a ranking (by Fav Num) higher than "Mike".
+  --
+  -- __NOTE:__ There are a few things to be aware about this statement.
+  --
+  --    * This only implements the full CASE statement, it does not
+  --      implement the "simple" CASE statement.
+  --
+  --
+  --    * At least one 'when_' and 'then_' is mandatory otherwise it will
+  --      emit an error.
+  --
+  --
+  --    * The 'else_' is also mandatory, unlike the SQL statement in which
+  --      if the @ELSE@ is omitted it will return a @NULL@. You can
+  --      reproduce this via 'nothing'.
+  --
+  -- /Since: 2.1.2/
+  case_ :: PersistField a => [(expr (Value Bool), expr (Value a))] -> expr (Value a) -> expr (Value a)
 
 -- Fixity declarations
 infixl 9 ^.
@@ -330,6 +392,23 @@ infix  4 ==., >=., >., <=., <., !=.
 infixr 3 &&., =., +=., -=., *=., /=.
 infixr 2 ||., `InnerJoin`, `CrossJoin`, `LeftOuterJoin`, `RightOuterJoin`, `FullOuterJoin`, `like`
 
+-- | Syntax sugar for 'case_'.
+--
+-- /Since: 2.1.2/
+when_ :: expr (Value Bool) -> () -> expr a -> (expr (Value Bool), expr a)
+when_ cond _ expr = (cond, expr)
+
+-- | Syntax sugar for 'case_'.
+--
+-- /Since: 2.1.2/
+then_ :: ()
+then_ = ()
+
+-- | Syntax sugar for 'case_'.
+--
+-- /Since: 2.1.2/
+else_ :: expr a -> expr a
+else_ = id
 
 -- | A single value (as opposed to a whole entity).  You may use
 -- @('^.')@ or @('?.')@ to get a 'Value' from an 'Entity'.
@@ -338,6 +417,11 @@ data Value a = Value a deriving (Eq, Ord, Show, Typeable)
 -- <https://ghc.haskell.org/trac/ghc/ticket/6124>
 
 
+-- | /Since: 1.4.4/
+instance Functor Value where
+  fmap f (Value a) = Value (f a)
+
+
 -- | Unwrap a 'Value'.
 --
 -- /Since: 1.4.1/
