@@ -7,6 +7,8 @@
            , MultiParamTypeClasses
            , OverloadedStrings
            , UndecidableInstances
+           , ScopedTypeVariables
+           , InstanceSigs
  #-}
 -- | This is an internal module, anything exported by this module
 -- may change without a major version bump.  Please use only
@@ -27,7 +29,9 @@ module Database.Esqueleto.Internal.Sql
   , insertSelectDistinct
   , insertSelect
     -- * The guts
+  , unsafeSqlCase
   , unsafeSqlBinOp
+  , unsafeSqlBinOpList
   , unsafeSqlValue
   , unsafeSqlFunction
   , unsafeSqlExtractSubField
@@ -48,15 +52,18 @@ module Database.Esqueleto.Internal.Sql
 import Control.Applicative (Applicative(..), (<$>), (<$))
 import Control.Arrow ((***), first)
 import Control.Exception (throw, throwIO)
-import Control.Monad ((>=>), ap, void, MonadPlus(..))
+import Control.Monad (ap, MonadPlus(..), liftM)
 import Control.Monad.IO.Class (MonadIO(..))
 import Control.Monad.Trans.Class (lift)
-import Control.Monad.Trans.Resource (MonadResource)
+import qualified Control.Monad.Trans.Reader as R
 import Data.Int (Int64)
 import Data.List (intersperse)
 import Data.Monoid (Monoid(..), (<>))
 import Data.Proxy (Proxy(..))
 import Database.Esqueleto.Internal.PersistentImport
+import Database.Persist.Sql.Util (
+    entityColumnNames, entityColumnCount, parseEntityValues, isIdField
+  , hasCompositeKey)
 import qualified Control.Monad.Trans.State as S
 import qualified Control.Monad.Trans.Writer as W
 import qualified Data.Conduit as C
@@ -65,6 +72,8 @@ import qualified Data.HashSet as HS
 import qualified Data.Text as T
 import qualified Data.Text.Lazy as TL
 import qualified Data.Text.Lazy.Builder as TLB
+import Data.Acquire (with, allocateAcquire, Acquire)
+import Control.Monad.Trans.Resource (MonadResource)
 
 import Database.Esqueleto.Internal.Language
 
@@ -111,7 +120,7 @@ instance Monoid SideData where
 
 -- | A part of a @FROM@ clause.
 data FromClause =
-    FromStart Ident (EntityDef SqlType)
+    FromStart Ident EntityDef
   | FromJoin FromClause JoinKind FromClause (Maybe (SqlExpr (Value Bool)))
   | OnClause (SqlExpr (Value Bool))
 
@@ -223,7 +232,7 @@ newIdentFor = Q . lift . try . unDBName
 
 
 -- | Information needed to escape and use identifiers.
-type IdentInfo = (Connection, IdentState)
+type IdentInfo = (SqlBackend, IdentState)
 
 
 -- | Use an identifier.
@@ -249,6 +258,9 @@ data SqlExpr a where
   -- interpolated by the SQL backend.
   ERaw     :: NeedParens -> (IdentInfo -> (TLB.Builder, [PersistValue])) -> SqlExpr (Value a)
 
+  -- Used to support non-id/composite primary keys
+  ERawList :: (IdentInfo -> ([TLB.Builder], [PersistValue])) -> SqlExpr (Value a)
+
   -- 'EList' and 'EEmptyList' are used by list operators.
   EList      :: SqlExpr (Value a) -> SqlExpr (ValueList a)
   EEmptyList :: SqlExpr (ValueList a)
@@ -330,33 +342,50 @@ instance Esqueleto SqlQuery SqlExpr SqlBackend where
   sub_select         = sub SELECT
   sub_selectDistinct = sub SELECT_DISTINCT
 
-  EEntity ident ^. field =
-    ERaw Never $ \info -> (useIdent info ident <> ("." <> fieldName info field), [])
+  (^.) :: forall val typ. (PersistEntity val, PersistField typ)
+       => SqlExpr (Entity val) -> EntityField val typ -> SqlExpr (Value typ)
+  EEntity ident ^. field
+    | isComposite = ERawList   $ \info -> (dot info <$> compositeFields pdef,  [])
+    | otherwise   = ERaw Never $ \info -> (dot info  $  persistFieldDef field, [])
+    where
+      isComposite = isIdField field && hasCompositeKey ed
+      dot info x  = useIdent info ident <> "." <> fromDBName info (fieldDB x)
+      ed          = entityDef $ getEntityVal (Proxy :: Proxy (SqlExpr (Entity val)))
+      Just pdef   = entityPrimary ed
 
   EMaybe r ?. field = maybelize (r ^. field)
     where
       maybelize :: SqlExpr (Value a) -> SqlExpr (Value (Maybe a))
-      maybelize (ERaw p f) = ERaw p f
+      maybelize (ERaw p f)   = ERaw p f
+      maybelize (ERawList f) = ERawList f
 
-  val = ERaw Never . const . (,) "?" . return . toPersistValue
+  val v =
+    case toPersistValue v of
+      PersistList vs -> ERawList   $ const (replicate (length vs) "?", vs)
+      pv             -> ERaw Never $ const ("?", [pv])
 
-  isNothing (ERaw p f) = ERaw Parens $ first ((<> " IS NULL") . parensM p) . f
-  just (ERaw p f) = ERaw p f
-  nothing   = unsafeSqlValue "NULL"
-  joinV (ERaw p f) = ERaw p f
+  isNothing (ERaw p f)   = ERaw Parens $ first ((<> " IS NULL") . parensM p) . f
+  isNothing (ERawList f) = ERaw Parens $ first (intersperseB " AND " . map (<> " IS NULL")) . f
+  just (ERaw p f)   = ERaw p f
+  just (ERawList f) = ERawList f
+  nothing = unsafeSqlValue "NULL"
+  joinV (ERaw p f)   = ERaw p f
+  joinV (ERawList f) = ERawList f
   countRows = unsafeSqlValue "COUNT(*)"
   count (ERaw _ f) = ERaw Never $ \info -> let (b, vals) = f info
                                            in ("COUNT" <> parens b, vals)
+  count (ERawList _) = unsafeSqlValue "COUNT(*)" -- Assumes no NULLs on a PK
 
   not_ (ERaw p f) = ERaw Never $ \info -> let (b, vals) = f info
                                           in ("NOT " <> parensM p b, vals)
+  not_ (ERawList _) = unexpectedCompositeKeyError
 
-  (==.) = unsafeSqlBinOp " = "
+  (==.) = unsafeSqlBinOpList " = " " AND "
+  (!=.) = unsafeSqlBinOpList " != " " OR "
   (>=.) = unsafeSqlBinOp " >= "
   (>.)  = unsafeSqlBinOp " > "
   (<=.) = unsafeSqlBinOp " <= "
   (<.)  = unsafeSqlBinOp " < "
-  (!=.) = unsafeSqlBinOp " != "
   (&&.) = unsafeSqlBinOp " AND "
   (||.) = unsafeSqlBinOp " OR "
   (+.)  = unsafeSqlBinOp " + "
@@ -365,13 +394,18 @@ instance Esqueleto SqlQuery SqlExpr SqlBackend where
   (*.)  = unsafeSqlBinOp " * "
 
   random_  = unsafeSqlValue "RANDOM()"
-  sum_     = unsafeSqlFunction "SUM"
   round_   = unsafeSqlFunction "ROUND"
   ceiling_ = unsafeSqlFunction "CEILING"
   floor_   = unsafeSqlFunction "FLOOR"
-  avg_     = unsafeSqlFunction "AVG"
+
+  sum_     = unsafeSqlFunction "SUM"
   min_     = unsafeSqlFunction "MIN"
   max_     = unsafeSqlFunction "MAX"
+  avg_     = unsafeSqlFunction "AVG"
+  lower_   = unsafeSqlFunction "LOWER"
+
+  coalesce              = unsafeSqlFunctionParens "COALESCE"
+  coalesceDefault exprs = unsafeSqlFunctionParens "COALESCE" . (exprs ++) . return . just
 
   like    = unsafeSqlBinOp    " LIKE "
   (%)     = unsafeSqlValue    "'%'"
@@ -401,13 +435,16 @@ instance Esqueleto SqlQuery SqlExpr SqlBackend where
   field *=. expr = setAux field (\ent -> ent ^. field *. expr)
   field /=. expr = setAux field (\ent -> ent ^. field /. expr)
 
-  (<#) _ (ERaw _ f) = EInsert Proxy f
+  (<#) _ (ERaw _ f)   = EInsert Proxy f
+  (<#) _ (ERawList _) = unexpectedCompositeKeyError
 
   (EInsert _ f) <&> (ERaw _ g) = EInsert Proxy $ \x ->
     let (fb, fv) = f x
         (gb, gv) = g x
     in (fb <> ", " <> gb, fv ++ gv)
+  (EInsert _ _) <&> (ERawList _) = unexpectedCompositeKeyError
 
+  case_ = unsafeSqlCase
 
 instance ToSomeValues SqlExpr (SqlExpr (Value a)) where
   toSomeValues a = [SomeValue a]
@@ -416,6 +453,7 @@ fieldName :: (PersistEntity val, PersistField typ)
           => IdentInfo -> EntityField val typ -> TLB.Builder
 fieldName info = fromDBName info . fieldDB . persistFieldDef
 
+-- FIXME: Composite/non-id pKS not supported on set
 setAux :: (PersistEntity val, PersistField typ)
        => EntityField val typ
        -> (SqlExpr (Entity val) -> SqlExpr (Value typ))
@@ -443,6 +481,31 @@ ifNotEmptyList (EList _)  _ x = x
 ----------------------------------------------------------------------
 
 
+-- | (Internal) Create a case statement.
+--
+-- Since: 2.1.1
+unsafeSqlCase :: PersistField a => [(SqlExpr (Value Bool), SqlExpr (Value a))] -> SqlExpr (Value a) -> SqlExpr (Value a)
+unsafeSqlCase when (ERaw p1 f1) = ERaw Never buildCase
+  where
+    buildCase :: IdentInfo -> (TLB.Builder, [PersistValue])
+    buildCase info =
+        let (b1, vals1) = f1 info
+            (b2, vals2) = mapWhen when info
+        in ( "CASE" <> b2 <> " ELSE " <> parensM p1 b1 <> " END", vals2 <> vals1)
+
+    mapWhen :: [(SqlExpr (Value Bool), SqlExpr (Value a))] -> IdentInfo -> (TLB.Builder, [PersistValue])
+    mapWhen []    _    = error "unsafeSqlCase: empty when list."
+    mapWhen when' info = foldl (foldHelp info) (mempty, mempty) when'
+
+    foldHelp :: IdentInfo -> (TLB.Builder, [PersistValue]) -> (SqlExpr (Value Bool), SqlExpr (Value a)) -> (TLB.Builder, [PersistValue])
+    foldHelp info (b0, vals0) (ERaw p1' f1', ERaw p2 f2) =
+        let (b1, vals1) = f1' info
+            (b2, vals2) = f2 info
+        in ( b0 <> " WHEN " <> parensM p1' b1 <> " THEN " <> parensM p2 b2, vals0 <> vals1 <> vals2 )
+    foldHelp _ _ _ = unexpectedCompositeKeyError
+unsafeSqlCase _ (ERawList _) = unexpectedCompositeKeyError
+
+
 -- | (Internal) Create a custom binary operator.  You /should/
 -- /not/ use this function directly since its type is very
 -- general, you should always use it with an explicit type
@@ -462,9 +525,36 @@ unsafeSqlBinOp op (ERaw p1 f1) (ERaw p2 f2) = ERaw Parens f
                  (b2, vals2) = f2 info
              in ( parensM p1 b1 <> op <> parensM p2 b2
                 , vals1 <> vals2 )
+unsafeSqlBinOp _ _ _ = unexpectedCompositeKeyError
 {-# INLINE unsafeSqlBinOp #-}
 
 
+-- | Similar to 'unsafeSqlBinOp', but may also be applied to
+-- composite keys.  Uses the operator given as the second
+-- argument whenever applied to composite keys.
+--
+-- Usage example:
+--
+-- @
+-- (==.) :: SqlExpr (Value a) -> SqlExpr (Value a) -> SqlExpr (Value Bool)
+-- (==.) = unsafeSqlBinOpList " = " " AND "
+-- @
+unsafeSqlBinOpList :: TLB.Builder -> TLB.Builder -> SqlExpr (Value a) -> SqlExpr (Value b) -> SqlExpr (Value c)
+unsafeSqlBinOpList op _ a@(ERaw _ _)    b@(ERaw _ _)  = unsafeSqlBinOp op a b
+unsafeSqlBinOpList op sep (ERawList f1) (ERawList f2) = ERaw Never f
+  where
+    f info = let (b1, vals1) = f1 info
+                 (b2, vals2) = f2 info
+             in ( intersperseB sep . map (\(a,b) -> a <> op <> b) $
+                  zip b1 b2
+                , vals1 <> vals2 )
+unsafeSqlBinOpList op sep a b = unsafeSqlBinOpList op sep (promote a) (promote b)
+  where
+    promote :: SqlExpr (Value d) -> SqlExpr (Value d)
+    promote (ERaw p f)     = ERawList $ first (return . parensM p) . f
+    promote x@(ERawList _) = x
+{-# INLINE unsafeSqlBinOpList #-}
+
 -- | (Internal) A raw SQL value.  The same warning from
 -- 'unsafeSqlBinOp' applies to this function as well.
 unsafeSqlValue :: TLB.Builder -> SqlExpr (Value a)
@@ -494,6 +584,16 @@ unsafeSqlExtractSubField subField arg =
           uncommas' $ map (\(ERaw _ f) -> f info) $ toArgList arg
     in ("EXTRACT" <> parens (subField <> " FROM " <> argsTLB), argsVals)
 
+-- | (Internal) A raw SQL function. Preserves parentheses around arguments.
+-- See 'unsafeSqlBinOp' for warnings.
+unsafeSqlFunctionParens :: UnsafeSqlFunctionArgument a =>
+                           TLB.Builder -> a -> SqlExpr (Value b)
+unsafeSqlFunctionParens name arg =
+  ERaw Never $ \info ->
+    let (argsTLB, argsVals) =
+          uncommas' $ map (\(ERaw p f) -> first (parensM p) (f info)) $ toArgList arg
+    in (name <> parens argsTLB, argsVals)
+
 
 class UnsafeSqlFunctionArgument a where
   toArgList :: a -> [SqlExpr (Value ())]
@@ -525,6 +625,7 @@ instance ( UnsafeSqlFunctionArgument a
 -- unless you know what you're doing!
 veryUnsafeCoerceSqlExprValue :: SqlExpr (Value a) -> SqlExpr (Value b)
 veryUnsafeCoerceSqlExprValue (ERaw p f) = ERaw p f
+veryUnsafeCoerceSqlExprValue (ERawList f) = ERawList f
 
 
 -- | (Internal) Coerce a value's type from 'SqlExpr (ValueList
@@ -541,19 +642,20 @@ veryUnsafeCoerceSqlExprValueList EEmptyList =
 -- | (Internal) Execute an @esqueleto@ @SELECT@ 'SqlQuery' inside
 -- @persistent@'s 'SqlPersistT' monad.
 rawSelectSource :: ( SqlSelect a r
-                   , MonadResource m
-                   , MonadSqlPersist m )
+                   , MonadIO m1
+                   , MonadIO m2 )
                  => Mode
                  -> SqlQuery a
-                 -> m (C.Source m r)
-rawSelectSource mode query = src
+                 -> SqlPersistT m1 (Acquire (C.Source m2 r))
+rawSelectSource mode query =
+      do
+        conn <- R.ask
+        res <- run conn
+        return $ (C.$= massage) `fmap` res
     where
-      src = do
-        conn <- askSqlConn
-        return $ run conn C.$= massage
 
       run conn =
-        uncurry rawQuery $
+        uncurry rawQueryRes $
         first builderToText $
         toRawSql mode (conn, initialIdentState) query
 
@@ -570,11 +672,14 @@ rawSelectSource mode query = src
 -- | Execute an @esqueleto@ @SELECT@ query inside @persistent@'s
 -- 'SqlPersistT' monad and return a 'C.Source' of rows.
 selectSource :: ( SqlSelect a r
-                , MonadResource m
-                , MonadSqlPersist m )
+                , MonadResource m )
              => SqlQuery a
-             -> m (C.Source m r)
-selectSource = rawSelectSource SELECT
+             -> C.Source (SqlPersistT m) r
+selectSource query = do
+    src <- lift $ do
+        res <- rawSelectSource SELECT query
+        fmap snd $ allocateAcquire res
+    src
 
 
 -- | Execute an @esqueleto@ @SELECT@ query inside @persistent@'s
@@ -619,10 +724,12 @@ selectSource = rawSelectSource SELECT
 -- function composition that the @p@ inside the query is of type
 -- @SqlExpr (Entity Person)@.
 select :: ( SqlSelect a r
-          , MonadResource m
-          , MonadSqlPersist m )
-       => SqlQuery a -> m [r]
-select = selectSource >=> runSource
+          , MonadIO m )
+       => SqlQuery a -> SqlPersistT m [r]
+select query = do
+    res <- rawSelectSource SELECT query
+    conn <- R.ask
+    liftIO $ with res $ flip R.runReaderT conn . runSource
 
 
 -- | Execute an @esqueleto@ @SELECT DISTINCT@ query inside
@@ -630,26 +737,31 @@ select = selectSource >=> runSource
 -- rows.
 selectDistinctSource
   :: ( SqlSelect a r
-     , MonadResource m
-     , MonadSqlPersist m )
+     , MonadResource m )
   => SqlQuery a
-  -> m (C.Source m r)
-selectDistinctSource = rawSelectSource SELECT_DISTINCT
+  -> C.Source (SqlPersistT m) r
+selectDistinctSource query = do
+    src <- lift $ do
+        res <- rawSelectSource SELECT_DISTINCT query
+        fmap snd $ allocateAcquire res
+    src
 
 
 -- | Execute an @esqueleto@ @SELECT DISTINCT@ query inside
 -- @persistent@'s 'SqlPersistT' monad and return a list of rows.
 selectDistinct :: ( SqlSelect a r
-                  , MonadResource m
-                  , MonadSqlPersist m )
-               => SqlQuery a -> m [r]
-selectDistinct = selectDistinctSource >=> runSource
+                  , MonadIO m )
+               => SqlQuery a -> SqlPersistT m [r]
+selectDistinct query = do
+    res <- rawSelectSource SELECT_DISTINCT query
+    conn <- R.ask
+    liftIO $ with res $ flip R.runReaderT conn . runSource
 
 
 -- | (Internal) Run a 'C.Source' of rows.
-runSource :: MonadResource m =>
-             C.Source m r
-          -> m [r]
+runSource :: Monad m =>
+             C.Source (SqlPersistT m) r
+          -> SqlPersistT m [r]
 runSource src = src C.$$ CL.consume
 
 
@@ -658,14 +770,12 @@ runSource src = src C.$$ CL.consume
 
 -- | (Internal) Execute an @esqueleto@ statement inside
 -- @persistent@'s 'SqlPersistT' monad.
-rawEsqueleto :: ( MonadResource m
-                , MonadSqlPersist m
-                , SqlSelect a r )
+rawEsqueleto :: ( MonadIO m, SqlSelect a r )
            => Mode
            -> SqlQuery a
-           -> m Int64
+           -> SqlPersistT m Int64
 rawEsqueleto mode query = do
-  conn <- askSqlConn
+  conn <- R.ask
   uncurry rawExecuteCount $
     first builderToText $
     toRawSql mode (conn, initialIdentState) query
@@ -693,18 +803,16 @@ rawEsqueleto mode query = do
 -- from $ \\(appointment :: SqlExpr (Entity Appointment)) ->
 -- return ()
 -- @
-delete :: ( MonadResource m
-          , MonadSqlPersist m )
+delete :: ( MonadIO m )
        => SqlQuery ()
-       -> m ()
-delete = void . deleteCount
+       -> SqlPersistT m ()
+delete = liftM (const ()) . deleteCount
 
 
 -- | Same as 'delete', but returns the number of rows affected.
-deleteCount :: ( MonadResource m
-               , MonadSqlPersist m )
+deleteCount :: ( MonadIO m )
             => SqlQuery ()
-            -> m Int64
+            -> SqlPersistT m Int64
 deleteCount = rawEsqueleto DELETE
 
 
@@ -718,44 +826,23 @@ deleteCount = rawEsqueleto DELETE
 -- @
 -- update $ \p -> do
 -- set p [ PersonAge =. just (val thisYear) -. p ^. PersonBorn ]
--- where_ $ isNull (p ^. PersonAge)
+-- where_ $ isNothing (p ^. PersonAge)
 -- @
-update :: ( MonadResource m
-          , MonadSqlPersist m
+update :: ( MonadIO m
           , SqlEntity val )
        => (SqlExpr (Entity val) -> SqlQuery ())
-       -> m ()
-update = void . updateCount
+       -> SqlPersistT m ()
+update = liftM (const ()) . updateCount
 
 
 -- | Same as 'update', but returns the number of rows affected.
-updateCount :: ( MonadResource m
-               , MonadSqlPersist m
+updateCount :: ( MonadIO m
                , SqlEntity val )
             => (SqlExpr (Entity val) -> SqlQuery ())
-            -> m Int64
+            -> SqlPersistT m Int64
 updateCount = rawEsqueleto UPDATE . from
 
 
--- | Insert a 'PersistField' for every selected value.
-insertSelect :: ( MonadResource m
-                , MonadSqlPersist m
-                , PersistEntity a )
-             => SqlQuery (SqlExpr (Insertion a)) -> m ()
-insertSelect =
-  void . rawEsqueleto (INSERT_INTO SELECT) . fmap EInsertFinal
-
-
--- | Insert a 'PersistField' for every unique selected value.
-insertSelectDistinct
-  :: ( MonadResource m
-     , MonadSqlPersist m
-     , PersistEntity a )
-  => SqlQuery (SqlExpr (Insertion a)) -> m ()
-insertSelectDistinct =
-  void . rawEsqueleto (INSERT_INTO SELECT_DISTINCT) . fmap EInsertFinal
-
-
 ----------------------------------------------------------------------
 
 
@@ -813,7 +900,10 @@ data Mode =
 
 
 uncommas :: [TLB.Builder] -> TLB.Builder
-uncommas = mconcat . intersperse ", " . filter (/= mempty)
+uncommas = intersperseB ", "
+
+intersperseB :: TLB.Builder -> [TLB.Builder] -> TLB.Builder
+intersperseB a = mconcat . intersperse a . filter (/= mempty)
 
 uncommas' :: Monoid a => [(TLB.Builder, a)] -> (TLB.Builder, a)
 uncommas' = (uncommas *** mconcat) . unzip
@@ -871,24 +961,30 @@ makeFrom info mode fs = ret
     fromKind RightOuterJoinKind = " RIGHT OUTER JOIN "
     fromKind FullOuterJoinKind  = " FULL OUTER JOIN "
 
-    makeOnClause (ERaw _ f) = first (" ON " <>) (f info)
+    makeOnClause (ERaw _ f)   = first (" ON " <>) (f info)
+    makeOnClause (ERawList _) = unexpectedCompositeKeyError
 
     mkExc :: SqlExpr (Value Bool) -> OnClauseWithoutMatchingJoinException
     mkExc (ERaw _ f) =
       OnClauseWithoutMatchingJoinException $
       TL.unpack $ TLB.toLazyText $ fst (f info)
+    mkExc (ERawList _) = unexpectedCompositeKeyError
+
+unexpectedCompositeKeyError :: a
+unexpectedCompositeKeyError = error "non-id/composite keys not expected here"
 
 
 makeSet :: IdentInfo -> [SetClause] -> (TLB.Builder, [PersistValue])
 makeSet _    [] = mempty
-makeSet info os = first ("\nSET " <>) $ uncommas' (map mk os)
+makeSet info os = first ("\nSET " <>) . uncommas' $ concatMap mk os
   where
-    mk (SetClause (ERaw _ f)) = f info
-
+    mk (SetClause (ERaw _ f))   = [f info]
+    mk (SetClause (ERawList _)) = error "esqueleto/makeSet: FIXME"
 
 makeWhere :: IdentInfo -> WhereClause -> (TLB.Builder, [PersistValue])
-makeWhere _    NoWhere            = mempty
-makeWhere info (Where (ERaw _ f)) = first ("\nWHERE " <>) (f info)
+makeWhere _    NoWhere              = mempty
+makeWhere info (Where (ERaw _ f))   = first ("\nWHERE " <>) (f info)
+makeWhere _    (Where (ERawList _)) = unexpectedCompositeKeyError
 
 
 makeGroupBy :: IdentInfo -> GroupByClause -> (TLB.Builder, [PersistValue])
@@ -898,22 +994,26 @@ makeGroupBy info (GroupBy fields) = first ("\nGROUP BY " <>) build
     build = uncommas' $ map (\(SomeValue (ERaw _ f)) -> f info) fields
 
 makeHaving :: IdentInfo -> WhereClause -> (TLB.Builder, [PersistValue])
-makeHaving _    NoWhere            = mempty
-makeHaving info (Where (ERaw _ f)) = first ("\nHAVING " <>) (f info)
+makeHaving _    NoWhere               = mempty
+makeHaving info (Where (ERaw _ f))    = first ("\nHAVING " <>) (f info)
+makeHaving _    (Where (ERawList _ )) = unexpectedCompositeKeyError
 
 makeOrderBy :: IdentInfo -> [OrderByClause] -> (TLB.Builder, [PersistValue])
 makeOrderBy _    [] = mempty
-makeOrderBy info os = first ("\nORDER BY " <>) $ uncommas' (map mk os)
+makeOrderBy info os = first ("\nORDER BY " <>) . uncommas' $ concatMap mk os
   where
-    mk :: OrderByClause -> (TLB.Builder, [PersistValue])
-    mk (EOrderBy t (ERaw p f)) = first ((<> orderByType t) . parensM p) (f info)
-    mk EOrderRandom = first ((<> "RANDOM()")) mempty
+    mk :: OrderByClause -> [(TLB.Builder, [PersistValue])]
+    mk (EOrderBy t (ERaw p f))   = [first ((<> orderByType t) . parensM p) (f info)]
+    mk (EOrderBy t (ERawList f)) = let fs = fst $ f info
+                                       vals = repeat []
+                                   in zip (map (<> orderByType t) fs) vals
+    mk EOrderRandom = [first ((<> "RANDOM()")) mempty]
     orderByType ASC  = " ASC"
     orderByType DESC = " DESC"
 
 
 makeLimit :: IdentInfo -> LimitClause -> [OrderByClause] -> (TLB.Builder, [PersistValue])
-makeLimit (conn,_) (Limit ml mo) orderByClauses =
+makeLimit (conn, _) (Limit ml mo) orderByClauses =
   let limitRaw = connLimitOffset conn (v ml, v mo) hasOrderClause "\n"
       hasOrderClause = not (null orderByClauses)
       v = maybe 0 fromIntegral
@@ -978,10 +1078,8 @@ instance PersistEntity a => SqlSelect (SqlExpr (Entity a)) (Entity a) where
   sqlSelectCols info expr@(EEntity ident) = ret
       where
         process ed = uncommas $
-                     map ((name <>) . fromDBName info) $
-                     (entityID ed:) $
-                     map fieldDB $
-                     entityFields ed
+                     map ((name <>) . TLB.fromText) $
+                     entityColumnNames ed (fst info)
         -- 'name' is the biggest difference between 'RawSql' and
         -- 'SqlSelect'.  We automatically create names for tables
         -- (since it's not the user who's writing the FROM
@@ -991,11 +1089,10 @@ instance PersistEntity a => SqlSelect (SqlExpr (Entity a)) (Entity a) where
         name = useIdent info ident <> "."
         ret = let ed = entityDef $ getEntityVal $ return expr
               in (process ed, mempty)
-  sqlSelectColCount = (+1) . length . entityFields . entityDef . getEntityVal
-  sqlSelectProcessRow (idCol:ent) =
-    Entity <$> fromPersistValue idCol
-           <*> fromPersistValues ent
-  sqlSelectProcessRow _ = Left "SqlSelect (Entity a): wrong number of columns."
+  sqlSelectColCount = entityColumnCount . entityDef . getEntityVal
+  sqlSelectProcessRow = parseEntityValues ed
+    where ed = entityDef $ getEntityVal $ (Proxy :: Proxy (SqlExpr (Entity a)))
+
 
 getEntityVal :: Proxy (SqlExpr (Entity a)) -> Proxy a
 getEntityVal = const Proxy
@@ -1018,6 +1115,10 @@ instance PersistEntity a => SqlSelect (SqlExpr (Maybe (Entity a))) (Maybe (Entit
 instance PersistField a => SqlSelect (SqlExpr (Value a)) (Value a) where
   sqlSelectCols info (ERaw p f) = let (b, vals) = f info
                                   in (parensM p b, vals)
+  sqlSelectCols info (ERawList f) = let (b, vals) = f info
+                                    in case b of
+                                         [b'] -> (parensM Parens b', vals)
+                                         _    -> unexpectedCompositeKeyError
   sqlSelectColCount = const 1
   sqlSelectProcessRow [pv] = Value <$> fromPersistValue pv
   sqlSelectProcessRow _    = Left "SqlSelect (Value a): wrong number of columns."
@@ -1512,3 +1613,21 @@ from16P = const Proxy
 
 to16 :: ((a,b),(c,d),(e,f),(g,h),(i,j),(k,l),(m,n),(o,p)) -> (a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p)
 to16 ((a,b),(c,d),(e,f),(g,h),(i,j),(k,l),(m,n),(o,p)) = (a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p)
+
+
+-- | Insert a 'PersistField' for every selected value.
+insertSelect :: (MonadIO m, PersistEntity a) =>
+  SqlQuery (SqlExpr (Insertion a)) -> SqlPersistT m ()
+insertSelect = insertGeneralSelect SELECT
+
+
+-- | Insert a 'PersistField' for every unique selected value.
+insertSelectDistinct :: (MonadIO m, PersistEntity a) =>
+  SqlQuery (SqlExpr (Insertion a)) -> SqlPersistT m ()
+insertSelectDistinct = insertGeneralSelect SELECT_DISTINCT
+
+
+insertGeneralSelect :: (MonadIO m, PersistEntity a) =>
+  Mode -> SqlQuery (SqlExpr (Insertion a)) -> SqlPersistT m ()
+insertGeneralSelect mode =
+  liftM (const ()) . rawEsqueleto (INSERT_INTO mode) . fmap EInsertFinal
