@@ -1,6 +1,8 @@
 {-# LANGUAGE ConstraintKinds
            , EmptyDataDecls
            , FlexibleContexts
+           , FlexibleInstances
+           , DeriveGeneric
            , GADTs
            , GeneralizedNewtypeDeriving
            , MultiParamTypeClasses
@@ -11,14 +13,17 @@
            , TypeFamilies
            , ScopedTypeVariables
            , CPP
+           , TypeSynonymInstances
  #-}
 module Main (main) where
 
 import Control.Applicative ((<$>))
+import Control.Exception (IOException)
 import Control.Monad (replicateM, replicateM_)
 import Control.Monad.IO.Class (MonadIO(liftIO))
 import Control.Monad.Logger (MonadLogger(..), runStderrLoggingT, runNoLoggingT)
 import Control.Monad.Trans.Control (MonadBaseControl(..))
+import Control.Monad.Trans.Reader (ReaderT)
 import Database.Esqueleto
 #if   defined (WITH_POSTGRESQL)
 import Database.Persist.Postgresql (withPostgresqlConn)
@@ -31,6 +36,9 @@ import Database.Persist.MySQL ( withMySQLConn
                               , defaultConnectInfo)
 #else
 import Database.Persist.Sqlite (withSqliteConn)
+#if MIN_VERSION_persistent_sqlite(2,1,3)
+import Database.Sqlite (SqliteException)
+#endif
 #endif
 import Database.Persist.TH
 import Test.Hspec
@@ -38,6 +46,7 @@ import Test.Hspec
 import qualified Control.Monad.Trans.Resource as R
 import qualified Data.Set as S
 import qualified Data.List as L
+import Data.Char (toLower, toUpper)
 
 
 -- Test schema
@@ -45,6 +54,8 @@ share [mkPersist sqlSettings, mkMigrate "migrateAll"] [persistUpperCase|
   Person
     name String
     age Int Maybe
+    weight Int Maybe
+    favNum Int
     deriving Eq Show
   BlogPost
     title String
@@ -54,6 +65,41 @@ share [mkPersist sqlSettings, mkMigrate "migrateAll"] [persistUpperCase|
     follower PersonId
     followed PersonId
     deriving Eq Show
+  Frontcover
+    number Int
+    title String
+    Primary number
+    deriving Eq Show
+  Article
+    title String
+    frontcoverNumber Int
+    Foreign Frontcover fkfrontcover frontcoverNumber
+    deriving Eq Show
+  Tag
+    name String
+    Primary name
+    deriving Eq Show
+  ArticleTag
+    articleId ArticleId
+    tagId     TagId
+    Primary   articleId tagId
+    deriving Eq Show
+  Article2
+    title String
+    frontcoverId FrontcoverId
+    deriving Eq Show
+  Point
+    x Int
+    y Int
+    name String
+    Primary x y
+    deriving Eq Show
+  Circle
+    centerX Int
+    centerY Int
+    name String
+    Foreign Point fkpoint centerX centerY
+    deriving Eq Show
 |]
 
 -- | this could be achieved with S.fromList, but not all lists
@@ -63,10 +109,11 @@ sameElementsAs l1 l2 = null (l1 L.\\ l2)
 
 main :: IO ()
 main = do
-  let p1 = Person "John"  (Just 36)
-      p2 = Person "Rachel" Nothing
-      p3 = Person "Mike"  (Just 17)
-      p4 = Person "Livia" (Just 17)
+  let p1 = Person "John"   (Just 36) Nothing   1
+      p2 = Person "Rachel" Nothing   (Just 37) 2
+      p3 = Person "Mike"   (Just 17) Nothing   3
+      p4 = Person "Livia"  (Just 17) (Just 18) 4
+      p5 = Person "Mitch"  Nothing   Nothing   5
   hspec $ do
     describe "select" $ do
       it "works for a single value" $
@@ -215,6 +262,38 @@ main = do
                   return p
           liftIO $ ret5 `shouldBe` [ p1e, p4e, p3e, p2e ]
 
+      it "works with non-id primary key" $
+        run $ do
+          let fc = Frontcover number ""
+              number = 101
+              Right thePk = keyFromValues [toPersistValue number]
+          fcPk <- insert fc
+          [Entity _ ret] <- select $ from $ return
+          liftIO $ do
+            ret `shouldBe` fc
+            fcPk `shouldBe` thePk
+
+      it "works when returning a custom non-composite primary key from a query" $
+        run $ do
+          let name = "foo"
+              t = Tag name
+              Right thePk = keyFromValues [toPersistValue name]
+          tagPk <- insert t
+          [Value ret] <- select $ from $ \t' -> return (t'^.TagId)
+          liftIO $ do
+            ret `shouldBe` thePk
+            thePk `shouldBe` tagPk
+
+      it "works when returning a composite primary key from a query" $
+        pendingWith "Need to refactor 'Value a's SqlQuery instance"
+        {-
+        run $ do
+          let p = Point 10 20 ""
+          thePk <- insert p
+          [Value ppk] <- select $ from $ \p' -> return (p'^.PointId)
+          liftIO $ ppk `shouldBe` thePk
+        -}
+
 
     describe "select/JOIN" $ do
       it "works with a LEFT OUTER JOIN" $
@@ -270,6 +349,96 @@ main = do
              return (p, mb)
         ) `shouldThrow` (\(OnClauseWithoutMatchingJoinException _) -> True)
 
+      it "works with ForeignKey to a non-id primary key returning one entity" $
+        run $ do
+          let fc = Frontcover number ""
+              article = Article "Esqueleto supports composite pks!" number
+              number = 101
+              Right thePk = keyFromValues [toPersistValue number]
+          fcPk <- insert fc
+          insert_ article
+          [Entity _ retFc] <- select $
+            from $ \(a `InnerJoin` f) -> do
+              on (f^.FrontcoverNumber ==. a^.ArticleFrontcoverNumber)
+              return f
+          liftIO $ do
+            retFc `shouldBe` fc
+            fcPk `shouldBe` thePk
+
+      it "works with a ForeignKey to a non-id primary key returning both entities" $
+        run $ do
+          let fc = Frontcover number ""
+              article = Article "Esqueleto supports composite pks!" number
+              number = 101
+              Right thePk = keyFromValues [toPersistValue number]
+          fcPk <- insert fc
+          insert_ article
+          [(Entity _ retFc, Entity _ retArt)] <- select $
+            from $ \(a `InnerJoin` f) -> do
+              on (f^.FrontcoverNumber ==. a^.ArticleFrontcoverNumber)
+              return (f, a)
+          liftIO $ do
+            retFc `shouldBe` fc
+            retArt `shouldBe` article
+            fcPk `shouldBe` thePk
+            articleFkfrontcover retArt `shouldBe` thePk
+
+      it "works with a non-id primary key returning one entity" $
+        run $ do
+          let fc = Frontcover number ""
+              article = Article2 "Esqueleto supports composite pks!" thePk
+              number = 101
+              Right thePk = keyFromValues [toPersistValue number]
+          fcPk <- insert fc
+          insert_ article
+          [Entity _ retFc] <- select $
+            from $ \(a `InnerJoin` f) -> do
+              on (f^.FrontcoverId ==. a^.Article2FrontcoverId)
+              return f
+          liftIO $ do
+            retFc `shouldBe` fc
+            fcPk `shouldBe` thePk
+
+      it "works with a composite primary key" $
+        pendingWith "Persistent does not create the CircleFkPoint constructor. See: https://github.com/yesodweb/persistent/issues/341"
+        {-
+        run $ do
+          let p = Point x y ""
+              c = Circle x y ""
+              x = 10
+              y = 15
+              Right thePk = keyFromValues [toPersistValue x, toPersistValue y]
+          pPk <- insert p
+          insert_ c
+          [Entity _ ret] <- select $ from $ \(c' `InnerJoin` p') -> do
+            on (p'^.PointId ==. c'^.CircleFkpoint)
+            return p'
+          liftIO $ do
+            ret `shouldBe` p
+            pPk `shouldBe` thePk
+       -}
+
+      it "works when joining via a non-id primary key" $
+        run $ do
+          let fc = Frontcover number ""
+              article = Article "Esqueleto supports composite pks!" number
+              tag = Tag "foo"
+              otherTag = Tag "ignored"
+              number = 101
+          insert_ fc
+          insert_ otherTag
+          artId <- insert article
+          tagId <- insert tag
+          insert_ $ ArticleTag artId tagId
+          [(Entity _ retArt, Entity _ retTag)] <- select $
+            from $ \(a `InnerJoin` at `InnerJoin` t) -> do
+              on (t^.TagId ==. at^.ArticleTagTagId)
+              on (a^.ArticleId ==. at^.ArticleTagArticleId)
+              return (a, t)
+          liftIO $ do
+            retArt `shouldBe` article
+            retTag `shouldBe` tag
+
     describe "select/where_" $ do
       it "works for a simple example with (==.)" $
         run $ do
@@ -365,6 +534,25 @@ main = do
                  return $ joinV $ max_ (p ^. PersonAge)
           liftIO $ ret `shouldBe` [ Value $ Just (36 :: Int) ]
 
+      it "works with lower_" $
+        run $ do
+          p1e <- insert' p1
+          p2e@(Entity _ bob) <- insert' $ Person "bob" (Just 36) Nothing   1
+
+          -- lower(name) == 'john'
+          ret <- select $
+                 from $ \p-> do
+                 where_ (lower_ (p ^. PersonName) ==. val (map toLower $ personName p1))
+                 return p
+          liftIO $ ret `shouldBe` [ p1e ]
+
+          -- name == lower('BOB')
+          ret <- select $
+                 from $ \p-> do
+                 where_ (p ^. PersonName ==. lower_ (val $ map toUpper $ personName bob))
+                 return p
+          liftIO $ ret `shouldBe` [ p2e ]
+
       it "works with random_" $
         run $ do
 #if defined(WITH_POSTGRESQL) || defined(WITH_MYSQL)
@@ -467,6 +655,20 @@ main = do
                                   , (p3e, Nothing,  Nothing)
                                   , (p2e, Just f21, Just p1e) ]
 
+      it "works with a composite primary key" $
+        run $ do
+          let p = Point x y ""
+              x = 10
+              y = 15
+              Right thePk = keyFromValues [toPersistValue x, toPersistValue y]
+          pPk <- insert p
+          [Entity _ ret] <- select $ from $ \p' -> do
+            where_ (p'^.PointId ==. val pPk)
+            return p'
+          liftIO $ do
+            ret `shouldBe` p
+            pPk `shouldBe` thePk
+
 
     describe "select/orderBy" $ do
       it "works with a single ASC field" $
@@ -546,6 +748,16 @@ main = do
           -- is 1/2^40, so this test should pass almost everytime.
           liftIO $ S.size rets `shouldSatisfy` (>2)
 
+      it "works on a composite primary key" $
+        run $ do
+          let ps = [Point 2 1 "", Point 1 2 ""]
+          mapM_ insert ps
+          eps <- select $
+            from $ \p' -> do
+              orderBy [asc (p'^.PointId)]
+              return p'
+          liftIO $ map entityVal eps `shouldBe` reverse ps
+
 
     describe "selectDistinct" $
       it "works on a simple example" $
@@ -560,6 +772,73 @@ main = do
                  return title
           liftIO $ ret `shouldBe` [ Value t1, Value t2, Value t3 ]
 
+    describe "coalesce/coalesceDefault" $ do
+      it "works on a simple example" $
+        run $ do
+          mapM_ insert' [p1, p2, p3, p4, p5]
+          ret1 <- select $
+                  from $ \p -> do
+                  orderBy [asc (p ^. PersonId)]
+                  return (coalesce [p ^. PersonAge, p ^. PersonWeight])
+          liftIO $ ret1 `shouldBe` [ Value (Just (36 :: Int))
+                                   , Value (Just 37)
+                                   , Value (Just 17)
+                                   , Value (Just 17)
+                                   , Value Nothing
+                                   ]
+
+          ret2 <- select $
+                  from $ \p -> do
+                  orderBy [asc (p ^. PersonId)]
+                  return (coalesceDefault [p ^. PersonAge, p ^. PersonWeight] (p ^. PersonFavNum))
+          liftIO $ ret2 `shouldBe` [ Value (36 :: Int)
+                                   , Value 37
+                                   , Value 17
+                                   , Value 17
+                                   , Value 5
+                                   ]
+
+      it "works with sub-queries" $
+        run $ do
+          p1id <- insert p1
+          p2id <- insert p2
+          p3id <- insert p3
+          _    <- insert p4
+          _    <- insert p5
+          _ <- insert $ BlogPost "a" p1id
+          _ <- insert $ BlogPost "b" p2id
+          _ <- insert $ BlogPost "c" p3id
+          ret <- select $
+                 from $ \b -> do
+                   let sub =
+                           from $ \p -> do
+                           where_ (p ^. PersonId ==. b ^. BlogPostAuthorId)
+                           return $ p ^. PersonAge
+                   return $ coalesceDefault [sub_select sub] (val (42 :: Int))
+          liftIO $ ret `shouldBe` [ Value (36 :: Int)
+                                  , Value 42
+                                  , Value 17
+                                  ]
+
+#if defined(WITH_POSTGRESQL) || defined(WITH_MYSQL)
+      it "works on PostgreSQL and MySQL with <2 arguments" $
+        run $ do
+          _ :: [Value (Maybe Int)] <- select $
+               from $ \p -> do
+               return (coalesce [p ^. PersonAge])
+          return True
+#else
+      it "throws an exception on SQLite with <2 arguments" $
+        run (select $
+             from $ \p -> do
+             return (coalesce [p ^. PersonAge]) :: SqlQuery (SqlExpr (Value (Maybe Int)))
+#if MIN_VERSION_persistent_sqlite(2,1,3)
+        ) `shouldThrow` (\(_ :: SqliteException) -> True)
+#else
+        ) `shouldThrow` (\(_ :: IOException) -> True)
+#endif
+#endif
+
     describe "text functions" $
       it "like, (%) and (++.) work on a simple example" $
          run $ do
@@ -620,21 +899,21 @@ main = do
           --             matched rows, not actually changed rows.
 #if   defined(WITH_POSTGRESQL)
           liftIO $ n `shouldBe` 2
-          liftIO $ ret `shouldBe` [ Entity p1k (Person anon (Just 73))
-                                  , Entity p2k (Person anon Nothing)
+          liftIO $ ret `shouldBe` [ Entity p1k (Person anon (Just 73) Nothing 1)
+                                  , Entity p2k (Person anon Nothing (Just 37) 2)
                                   , Entity p3k p3 ]
           -- MySQL: nulls appear first, and update returns actual number
           --        of changed rows
 #elif defined(WITH_MYSQL)
           liftIO $ n `shouldBe` 1
-          liftIO $ ret `shouldBe` [ Entity p2k (Person anon Nothing)
-                                  , Entity p1k (Person anon (Just 73))
+          liftIO $ ret `shouldBe` [ Entity p2k (Person anon Nothing (Just 37) 2)
+                                  , Entity p1k (Person anon (Just 73) Nothing 1)
                                   , Entity p3k p3 ]
 #else
           -- SQLite: nulls appear first, update returns matched rows.
           liftIO $ n `shouldBe` 2
-          liftIO $ ret `shouldBe` [ Entity p2k (Person anon Nothing)
-                                  , Entity p1k (Person anon (Just 73))
+          liftIO $ ret `shouldBe` [ Entity p2k (Person anon Nothing (Just 37) 2)
+                                  , Entity p1k (Person anon (Just 73) Nothing 1)
                                   , Entity p3k p3 ]
 #endif
 
@@ -659,6 +938,24 @@ main = do
                                   , Entity p3k p3 { personAge = Just 7 }
                                   , Entity p2k p2 { personAge = Just 0 } ]
 
+      it "works with a composite primary key" $
+        pendingWith "Need refactor to support composite pks on ESet"
+        {-
+        run $ do
+          let p = Point x y ""
+              x = 10
+              y = 15
+              newX = 20
+              newY = 25
+              Right newPk = keyFromValues [toPersistValue newX, toPersistValue newY]
+          insert_ p
+          () <- update $ \p' -> do
+                set p' [PointId =. val newPk]
+          [Entity _ ret] <- select $ from $ return
+          liftIO $ do
+            ret `shouldBe` Point newX newY []
+        -}
+
       it "GROUP BY works with COUNT" $
         run $ do
           p1k <- insert p1
@@ -803,10 +1100,10 @@ main = do
             _ <- insert p2
             _ <- insert p3
             _ <- insert p4
-            _ <- insert $ Person "Jane"  Nothing
-            _ <- insert $ Person "Mark"  Nothing
-            _ <- insert $ Person "Sarah" Nothing
-            insert $ Person "Paul"  Nothing
+            _ <- insert $ Person "Jane"  Nothing Nothing 0
+            _ <- insert $ Person "Mark"  Nothing Nothing 0
+            _ <- insert $ Person "Sarah" Nothing Nothing 0
+            insert $ Person "Paul"  Nothing Nothing 0
           ret1 <- fmap (map unValue) $ select $ from $ \p -> do
                     orderBy [rand]
                     return (p ^. PersonId)
@@ -816,14 +1113,61 @@ main = do
 
           liftIO $ (ret1 == ret2) `shouldBe` False
 
+    describe "case" $ do
+      it "Works for a simple value based when - False" $
+        run $ do
+          ret <- select $
+            return $
+              case_
+                [ when_ (val False) then_ (val (1 :: Int)) ]
+                (else_ (val 2))
+
+          liftIO $ ret `shouldBe` [ Value 2 ]
+
+      it "Works for a simple value based when - True" $
+        run $ do
+          ret <- select $
+            return $
+              case_
+                [ when_ (val True) then_ (val (1 :: Int)) ]
+                (else_ (val 2))
+
+          liftIO $ ret `shouldBe` [ Value 1 ]
+
+      it "works for a semi-complicated query" $
+        run $ do
+          _ <- insert p1
+          _ <- insert p2
+          _ <- insert p3
+          _ <- insert p4
+          _ <- insert p5
+          ret <- select $
+            return $
+              case_
+                [ when_
+                    (exists $ from $ \p -> do
+                        where_ (p ^. PersonName ==. val "Mike"))
+                  then_
+                    (sub_select $ from $ \v -> do
+                        let sub =
+                                from $ \c -> do
+                                where_ (c ^. PersonName ==. val "Mike")
+                                return (c ^. PersonFavNum)
+                        where_ (v ^. PersonFavNum >. sub_select sub)
+                        return $ count (v ^. PersonName) +. val (1 :: Int)) ]
+                (else_ $ val (-1))
+
+          liftIO $ ret `shouldBe` [ Value (3) ]
+
+
 ----------------------------------------------------------------------
 
 
 insert' :: ( Functor m
-           , PersistStore m
-           , PersistMonadBackend m ~ PersistEntityBackend val
+           , PersistStore (PersistEntityBackend val)
+           , MonadIO m
            , PersistEntity val )
-        => val -> m (Entity val)
+        => val -> ReaderT (PersistEntityBackend val) m (Entity val)
 insert' v = flip Entity v <$> insert v
 
 
